//
//  BFRecordScreenCameraManager.swift
//  BFRecordScreenKit
//
//  Created by 胡志强 on 2022/1/13.
//

/*
 
 时长是根据摄像头文件时长相加得来的，预览/合成时需要考虑录音的时长问题
 
 videosticker 已经存储好了，在合成时，需要更新下进入的时间就可以了，不需要再切割了
 
 
 
 */

import Foundation
import GPUImage
import BFCommonKit
import BFUIKit
import BFMediaKit
import CoreMedia
import UIKit


let vpath = recordVideosDirectory + "camera_writer.mov"
let videoSize = CGSize(width: cScreenWidth, height: cScreenWidth * 16 / 9.0)

class BFRecordScreenCameraManager : BFRecordScreenBaseManager{
    //
    var rendView = GPUImageView()
    var cameraProgressV: BFCameraProgressView?
    
    // 录制完成回调
    var hasInitCallBack : (()->Void)?
    var recordEndCallBack : recordEndCallBack?
    var recordProgressCallBack : ((CMTime) -> Void)?
    
    var startTime = Date()
    var videoModel = PQEditVisionTrackMaterialsModel()
    
    // 用于打开摄像头初始流程
    static var firstOpenCamera : Bool = true
    //当前录制的缩略图数量
    var currVideoThumImagesNum = 0
    //当前摄像的序号
    var currVideoIndex = 0
    
    var currPlayTime : CMTime = .zero
    
    fileprivate var timerr:Timer?
   
    var movieWrite : GPUImageMovieWriter?
    
    lazy var camera : GPUImageStillCamera? = {
        let camera = GPUImageStillCamera(sessionPreset: AVCaptureSession.Preset.hd1920x1080.rawValue, cameraPosition: AVCaptureDevice.Position.back)
        camera?.outputImageOrientation = UIInterfaceOrientation.portrait
        camera?.delegate = self
        return camera
        
    }()
    
    var avplayerTimeObserver: NSKeyValueObservation?
    
    lazy var avplayer : AVPlayer = {
        let p = AVPlayer(playerItem: nil)
        avplayerTimeObserver = p.addPeriodicTimeObserver(forInterval: CMTime(value: 1, timescale: 100), queue: DispatchQueue.global(), using: {[weak self] currTime in
            
            guard let wself = self else { return }
            
            wself.currentAssetProgress = CMTime(seconds: (currTime + wself.currPlayTime).seconds, preferredTimescale: 1000)
            wself.dele?.currentAssetProgress = wself.currentAssetProgress
//            BFLog(1, message: "camera currTime: \(currTime.seconds)")
            DispatchQueue.main.async {[weak self] in
                guard let wself = self else { return }
                wself.dele?.updateSubtitle(time: wself.currentAssetProgress)
                wself.locationTo(time: wself.currentAssetProgress)
            }
            
        }) as? NSKeyValueObservation
        
        
        return p
    }()
    
    lazy var playerLayer : AVPlayerLayer = {
        let layer = AVPlayerLayer(player: avplayer)
        layer.frame = UIScreen.main.bounds
        return layer
    }()
    
// MARK: 重放
    
    deinit {
        if avplayerTimeObserver != nil {
            avplayer.removeTimeObserver(avplayerTimeObserver)
            avplayerTimeObserver?.invalidate()
        }
        
    }
    
    override func resetEnv(){
        guard let camera = camera else {
            cShowHUB(superView: nil, msg: "option_fail_camera".BFLocale)
            return
        }
        cameraProgressV?.recordItem = recordItem
        cameraProgressV?.collectionV.reloadData()
        
        initplayer()

        if let count = recordItem?.voiceStickers.count, count > 0 {
            currentAssetProgress = recordItem!.materialDuraion
            currVideoIndex = count
            cameraProgressV?.isHidden = false
            progreddL?.isHidden = false
            progreddL?.text = CMTimeGetSeconds(currentAssetProgress).formatDurationToHMS()
            cameraProgressV?.collectionV.scrollRectToVisible(CGRect(x: (cameraProgressV?.collectionV.contentSize.width)! - 1, y: (cameraProgressV?.collectionV.contentSize.height)! - 1, width: 1, height: 1), animated: false)


        }else {
            currVideoIndex = 0
            cameraProgressV?.isHidden = true
            progreddL?.isHidden = true
        }
        
        camera.startCapture()
        initerlizeWriter()

        camera.addTarget(filter)
        filter.addTarget(rendView)
        
        if BFRecordScreenCameraManager.firstOpenCamera {
            startRecord()
            DispatchQueue.main.asyncAfter(deadline: .now() + 0.3, execute: { [weak self] in
                guard let wself = self else { return }

                wself.endRecord()
                BFRecordScreenCameraManager.firstOpenCamera = false
            })
        }
        
        super.resetEnv()
    }
    
    func initplayer(){
        
        guard let preView = playView else {
            return
        }
                
        preView.layer.addSublayer(playerLayer)
        
    }
    
    override func startRecord(){
        guard let movieWrite = movieWrite else {
            cShowHUB(superView: nil, msg: "option_fail_record".BFLocale)
            return
        }
        
        playerLayer.removeFromSuperlayer()

        super.startRecord()

        currVideoThumImagesNum = 0
        startTime = Date()
        
        videoModel = PQEditVisionTrackMaterialsModel()
        videoModel.thumImgs = [UIImage]()
        videoModel.timelineCMIn = currentAssetProgress
        videoModel.timelineCMOut = currentAssetProgress
        currVideoIndex = recordItem?.videoStickers.count ?? 0
        
        // 为了第一时间能更新collection view数据
        if !BFRecordScreenCameraManager.firstOpenCamera{
            recordItem?.videoStickers.append(videoModel)
            progreddL?.isHidden = false
            cameraProgressV?.isHidden = false
        }
        movieWrite.startRecording()
        
        timerr = Timer.scheduledTimer(withTimeInterval: 0.01, repeats: true, block: {[weak self] _ in
            guard let wself = self else { return }
            if BFRecordScreenCameraManager.firstOpenCamera {
                return
            }
            //MARK: 进度计时器

            let currDur = wself.videoModel.timelineCMIn + movieWrite.duration
            wself.currentAssetProgress = currDur
            wself.videoModel.timelineCMOut = currDur
            wself.recordProgressCallBack?(currDur)
            wself.updateUI(progress: currDur)
//            BFLog(1, message: "currTime: \(wself.currentAssetProgress.seconds)")
        })
        RunLoop.current.add(timerr!, forMode: .common)
    }
    
    override func endRecord(){
        super.endRecord()
        
        timerr?.invalidate()
        timerr = nil
//        BFLog(1, message: " 拍摄时长 \(Date().timeIntervalSince(startTime))")
        let su = BFRecordScreenCameraManager.firstOpenCamera
        movieWrite?.finishRecording(completionHandler: {[weak self] in
            guard let wself = self else { return }
            
            if FileManager.default.fileExists(atPath: vpath) {
                if !su {
                    let newPath:String? = PQBridgeObject.p_setupFileRename(vpath)
                    if let finalPath = newPath, finalPath.count > 0 {
                        let dur = AVURLAsset(url: URL(fileURLWithPath: finalPath))
                        wself.videoModel.timelineCMOut = wself.videoModel.timelineCMIn + CMTime(seconds: dur.duration.seconds, preferredTimescale: 1000)
                        BFLog(1, message: "拍摄文件时长：\(wself.videoModel.timelineCMIn.seconds)~\(wself.videoModel.timelineCMOut.seconds),  \(dur.duration.seconds)")
                        wself.videoModel.locationPath = finalPath
                        wself.currentAssetProgress = wself.videoModel.timelineCMOut
                        wself.recordEndCallBack?(true, wself.videoModel)

                    }else{
                        // 录制失败后恢复原样
                        wself.revertLast()
                        cShowHUB(superView: nil, msg: "option_fail_file".BFLocale)
                        wself.recordEndCallBack?(false, nil)

                    }
                    DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) {[weak self] in
                        guard let wself = self else { return }
                        
                        wself.updateUI(progress:  wself.currentAssetProgress)
                    }

                }else{
                    wself.hasInitCallBack?()
                }
            }
            
            // 重置writer，提前为下次录制做准备
            wself.initerlizeWriter()
        })
    }
    
    override func cancelRecord(){
        
        revertLast()
        BFLog(1, message: "摄像录制小于1s")
        
        DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) {[weak self] in
            guard let wself = self else { return }
            
            wself.updateUI(progress:  wself.currentAssetProgress)
        }
    }
    
    // 录制失败 ， 不足一秒时恢复上一次状态
    func revertLast(){
        videoModel.timelineCMOut = videoModel.timelineCMIn
        currentAssetProgress = videoModel.timelineCMOut
        recordItem?.thumbImgs.removeAll(where: { img in
            videoModel.thumImgs?.contains(img) ?? false
        })
        DispatchQueue.main.async {[weak self] in
            guard let wself = self else { return }
            
            wself.cameraProgressV?.collectionV.reloadData()
        }
        videoModel.thumImgs?.removeAll()
        recordItem?.videoStickers.removeLast()
        
    }
    
    // 使用avplayer进行播放视频和音频
    override func play() {
        guard let preView = playView else {
            return
        }
        if playerLayer.superlayer == nil {
            preView.layer.addSublayer(playerLayer)
        }

        camera?.pauseCapture()
        
        // 100ms以内的重头开始播放
        if (recordItem?.materialDuraion.seconds)! - currentAssetProgress.seconds < 0.1 {
            currentAssetProgress = .zero
            currPlayTime = .zero
            locationTo(time: currentAssetProgress)
            avplayer.replaceCurrentItem(with: nil)
        }
        
        if avplayer.currentItem != nil {
            avplayer.play()
            recordPlayer?.play()
        }else{
            prepareToPlayNext()
        }
        
        
    }
    
    func playRecordVoice(_ tt:CMTime, needPlay:Bool = true){
        // 播放音频
        if let mod = recordItem?.voiceStickers.first(where: { m in
            CMTimeCompare(m.startCMTime, currentAssetProgress + CMTime(seconds: 0.033, preferredTimescale: 1000)) <= 0 && CMTimeCompare(m.endCMTime, currentAssetProgress) > 0
        }){
            recordPlayer?.replaceCurrentItem(with: AVPlayerItem(url: URL(fileURLWithPath: mod.wavFilePath)))
            
            BFLog(1, message: "record - tt: \(tt.seconds), curr:\(currentAssetProgress.seconds)")
            recordPlayer?.seek(to: tt, toleranceBefore: CMTime(seconds: 1, preferredTimescale: 1000), toleranceAfter: CMTime(seconds: 1, preferredTimescale: 1000)) { [weak recordPlayer] isFinished in
                if isFinished && needPlay{
                    recordPlayer?.play()
                }
            }
        }
    }
    
    
    // 定位到某个时间点，相应进度条的UI更新
    func locationTo(time:CMTime) {
        var toTime = CMTime(seconds: time.seconds, preferredTimescale: 1000)
        if CMTimeCompare(time, recordItem!.materialDuraion) > 0 {
            toTime = recordItem!.materialDuraion
            currentAssetProgress = toTime
        }
        progreddL?.text = CMTimeGetSeconds(currentAssetProgress).formatDurationToHMS()

        cameraProgressV?.collectionV.contentOffset = CGPoint(x: currentAssetProgress.seconds * 70.0 / 5.0, y: 0)
        
    }
    override func deleteRecord(at: CMTime) {
        if let videoS = recordItem?.videoStickers.enumerated().first(where: { mod in
            CMTimeCompare(mod.1.timelineCMIn, at) <= 0 && CMTimeCompare(mod.1.timelineCMOut, at) >= 0
        }), let item = recordItem {
            
            item.thumbImgs.removeAll(where: { img in
                videoS.1.thumImgs?.contains(img) ?? false
            })
            item.videoStickers.remove(at: videoS.0)
            
            let currDuration = videoS.1.timelineCMOut - videoS.1.timelineCMIn
            item.materialDuraion = item.materialDuraion - currDuration
            currentAssetProgress = videoS.1.timelineCMIn
            
            // 更新音频的时间点
            for model in item.voiceStickers {
                if CMTimeCompare(currentAssetProgress, model.startCMTime) < 0 {
                    model.startCMTime = model.startCMTime - currDuration
                    model.endCMTime = model.endCMTime - currDuration
                    BFLog(1, message: "camera删除后更新voice time：\(model.startCMTime.seconds) ~ \(model.endCMTime.seconds)")
                    item.titleStickers.forEach { tm in
                        if tm.recordId == model.recordId {
                            tm.timelineIn = tm.timelineIn - currDuration
                            tm.timelineOut = tm.timelineOut - currDuration
                        }
                    }
                }
            }
            // 更新录像的时间点
            for model in item.videoStickers {
                if CMTimeCompare(currentAssetProgress, model.timelineCMIn) < 0 {
                    model.timelineCMIn =  model.timelineCMIn - currDuration
                    model.timelineCMOut = model.timelineCMOut - currDuration
                    BFLog(1, message: "camera删除后更新video time：\(model.timelineCMIn.seconds) ~ \(model.timelineCMOut.seconds)")

                }
            }
        }
    }

    override func changeRecordMaterail(){
        //
        
        if recordItem?.videoStickers.count ?? 0 == 0 {
            currentAssetProgress = .zero
            resetEnv()
        } else {
            cameraProgressV?.collectionV.reloadData()
            cameraProgressV?.collectionV.contentOffset = CGPoint(x: currentAssetProgress.seconds * 70.0 / 5.0, y: 0)
            progreddL?.text = CMTimeGetSeconds(currentAssetProgress).formatDurationToHMS()

            cameraProgressV?.isHidden = false
            
            if currentAssetProgress == .zero {
                // 按钮为删除
            }
        }
    }
    
    override func pause() {
        avplayer.pause()
        recordPlayer?.pause()
        if currentAssetProgress == recordItem?.materialDuraion {
            camera?.resumeCameraCapture()
        }
    }
    
    override func dragEnd() {
        // 播放视频
        if let mod = recordItem?.videoStickers.first(where: { m in
            CMTimeCompare(m.timelineCMIn, currentAssetProgress) <= 0 && CMTimeCompare(m.timelineCMOut, currentAssetProgress) > 0
        }){
            avplayerReplaceItem(newItem: AVPlayerItem(url: URL(fileURLWithPath: mod.locationPath)))
            currPlayTime = mod.timelineCMIn

            let tt = currentAssetProgress - currPlayTime
            BFLog(1, message: "asset - tt: \(tt.seconds), curr:\(currentAssetProgress.seconds)")
            avplayer.seek(to: tt, toleranceBefore: CMTime(seconds: 1, preferredTimescale: 1000), toleranceAfter: CMTime(seconds: 1, preferredTimescale: 1000))

            playRecordVoice(tt, needPlay: false)
        }
    }
    
    func prepareToPlayNext(){
        if let mod = recordItem?.videoStickers.sorted(by: { m1, m2 in
            m1.timelineCMIn.seconds < m2.timelineCMIn.seconds
        }).first(where: { mod in
            CMTimeCompare(mod.timelineCMIn, currentAssetProgress + CMTime(seconds: 0.033, preferredTimescale: 1000)) <= 0 && CMTimeCompare(currentAssetProgress, mod.timelineCMOut) < 0
        }){
            currPlayTime = mod.timelineCMIn
            locationTo(time: mod.timelineCMIn)
            avplayerReplaceItem(newItem: AVPlayerItem(url: URL(fileURLWithPath: mod.locationPath)))
            avplayer.play()
            playRecordVoice(.zero)
        }
    }
    
    func avplayerReplaceItem(newItem: AVPlayerItem){
        if let item = avplayer.currentItem {
            NotificationCenter.default.removeObserver(self as Any, name: .AVPlayerItemDidPlayToEndTime, object: item)
        }
        
        avplayer.replaceCurrentItem(with: newItem)
        NotificationCenter.default.addObserver(forName: .AVPlayerItemDidPlayToEndTime, object: avplayer.currentItem, queue: .main) { [weak self] notice in
            guard let wself = self else {
                return
            }
            
            BFLog(1, message: "endPlay:\(wself.currentAssetProgress.seconds), 总时长:\(wself.recordItem!.materialDuraion.seconds), currplayer:\(wself.avplayer.currentItem?.currentTime().seconds)")
    
            // 如果离结束大于0.7s，代表还有下一段视频
            if ((wself.recordItem?.materialDuraion ?? .zero) - wself.currentAssetProgress).seconds > 0.7 {
                wself.prepareToPlayNext()
            }else{
                // 播放到末尾了
                wself.dele?.isEndPlay = true
                wself.dele?.pause()
            }
        }
    }
    
    override func clean() {
        filter.removeTarget(movieWrite)
        filter.removeTarget(rendView)
        camera?.removeTarget(filter)
        super.clean()
    }
    
    // MARK: - 私有方法
    func updateUI(progress:CMTime){
        DispatchQueue.main.async {[weak self] in
            guard let wself = self else { return }
            
            wself.progreddL?.text = CMTimeGetSeconds(progress).formatDurationToHMS()
            if wself.recordItem?.thumbImgs.count ?? 0 > 0 {
                wself.cameraProgressV?.updateCellWidth(index: wself.recordItem!.thumbImgs.count - 1, progress: progress)
            }
        }
    }
    
    // 添加新的缩略图
    func addNewThumb(){
        if let progressThumV = cameraProgressV {
//            progressThumV.collectionV.reloadData()
            progressThumV.addThumbImages()
        }
    }
    
    // 初始化 movieWrite
    func initerlizeWriter(){
        if FileManager.default.fileExists(atPath: vpath){
            try? FileManager.default.removeItem(atPath: vpath)
        }
        if movieWrite != nil {
            filter.removeTarget(movieWrite)
            movieWrite = nil
        }
        
        movieWrite = GPUImageMovieWriter(movieURL: URL(fileURLWithPath: vpath), size: videoSize)
        filter.addTarget(movieWrite)

    }
    
}

extension BFRecordScreenCameraManager : GPUImageVideoCameraDelegate {
    func willOutputSampleBuffer(_ sampleBuffer: CMSampleBuffer!) {
        if !recording || BFRecordScreenCameraManager.firstOpenCamera {
            return
        }
        let dur = CMTime(seconds: movieWrite?.duration.seconds ?? 0, preferredTimescale: 1000)
        // 获取缩略图
        let coculationCount = Int(dur.seconds / 5.0)
        if coculationCount >= currVideoThumImagesNum{
            if let img = imageFromBuffer(sampleBuffer){
                BFLog(1, message: "取一张缩略图出来 \(dur.seconds)， \(recordItem?.thumbImgs.count ?? 0)")
                recordItem?.thumbImgs.append(img)
                videoModel.thumImgs?.append(img)
                currVideoThumImagesNum += 1
                addNewThumb()
            }
        }
    }
    
    func imageFromBuffer(_ buff: CMSampleBuffer) -> UIImage? {
        if let buffer = CMSampleBufferGetImageBuffer(buff) {
            let ciImage = CIImage(cvPixelBuffer: buffer)

            return UIImage(ciImage: ciImage.oriented(forExifOrientation: 6),
                           scale: cScreenWidth / 50.0,
                           orientation: .up)

        }

        return nil
    }
    
}
